module data_mod

  use genrev3d_types
  use data_types
  use wave_types
  use rev_extrap_types
  use slow_types


  implicit none

  integer, external ::  sreed,srite,sseek_block
  logical, private :: adj,add
  integer, private :: n_bytes_io,i_ztau_rec,i_x_0_rec,i_y_0_rec

  real, allocatable, dimension (:,:,:,:), private :: data_io

contains

  !-----------------------------------------------------------------------------
  subroutine init_io_data(gen_par,data_par,wave_par)

    type (genrev3d_param_type) gen_par
    type (data_param_type) data_par
    type (wave_param_type) wave_par

    integer ierr

    !+++
    if_comp_domain_data_1 : if (gen_par%comp_domain(1:4) == 'midp') then

      n_bytes_io=data_par%n_mx*data_par%n_my* &
                 data_par%n_hx*data_par%n_hy* &
                 data_par%esize

!      i_x_0_rec=1
!      i_y_0_rec=1

      i_x_0_rec=(data_par%mx_0-wave_par%mx_0)/wave_par%d_mx + 1
      i_y_0_rec=(data_par%my_0-wave_par%my_0)/wave_par%d_my + 1

      write(0,*)'i_x_0_rec=',i_x_0_rec
      write(0,*)'i_y_0_rec=',i_y_0_rec

      allocate (data_io(data_par%n_mx,data_par%n_my, &
                        data_par%n_hx,data_par%n_hy), &
               stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of data_io")
      if(gen_par%debug2) write(0,*)'Allocate data_io with shape=',shape(data_io)

    else if_comp_domain_data_1

      n_bytes_io=data_par%n_off_x*data_par%n_off_y* &
                 data_par%esize
      i_x_0_rec=(data_par%off_x_0-wave_par%off_x_0)/wave_par%d_mx + 1
      i_y_0_rec=(data_par%off_y_0-wave_par%off_y_0)/wave_par%d_my + 1
  
      write(0,*)'i_x_0_rec=',i_x_0_rec
      write(0,*)'i_y_0_rec=',i_y_0_rec

      allocate (data_io(data_par%n_off_x,data_par%n_off_y, &
                        1,1), &
               stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of data_io")
      if(gen_par%debug2) write(0,*)'Allocate data_io with shape=',shape(data_io)

    end if if_comp_domain_data_1

   write(0,*)'data n_bytes_io=',n_bytes_io

    if(gen_par%tau_domain) then
      i_ztau_rec=(gen_par%rec_depth-wave_par%tau_0)/wave_par%d_tau+1
    else
      i_ztau_rec=(gen_par%rec_depth-wave_par%z_0)/wave_par%d_z+1
    end if

   write(0,*)'i_ztau_rec =',i_ztau_rec


  end subroutine init_io_data


  !-----------------------------------------------------------------------------
  subroutine output_data(tag_data,gen_par,data_par,wave_par,rev_extrap_par, &
                         wave_block)

   character(len=128) :: tag_data
   type (genrev3d_param_type) gen_par
   type (data_param_type) data_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:,:,:)   :: wave_block

   integer :: len_tag,i_t0,i_t_data,ret_sseek,nsseek_blocks

   len_tag=len_trim(tag_data)

   i_t0=wave_par%i_t_step(1)
   i_t_data=nint((rev_extrap_par%t-data_par%t_0)/data_par%d_t)+1

   if(i_t_data < 1) then
      write(0,*)'i_t_data=',i_t_data
     call seperr("Writing for data before origin")
   end if
   if(i_t_data > data_par%n_t ) then
     write(0,*)'i_t_data=',i_t_data
     call seperr("Writing data after end")
   end if

   nsseek_blocks=(i_t_data-1) +  &
                 ((data_par%i_sx-1)+(data_par%i_sy-1)*data_par%n_sx)*data_par%n_t

   write(0,*)'i_t_data=',i_t_data
   write(0,*)'nsseek_blocks=',nsseek_blocks

   ret_sseek=sseek_block(trim(tag_data),nsseek_blocks,n_bytes_io,0) 
   if (ret_sseek .ne. nsseek_blocks) then
   !{
     write(0,*)'ret_sseek=',ret_sseek
     write(0,*)'tag_data=',trim(tag_data)
     call seperr("Error block seeking data")
   !}
   end if


   if_comp_domain_data_1 : if (gen_par%comp_domain(1:4) == 'midp') then
     if(data_par%n_my > 1) then
       data_io(1:data_par%n_mx,1:data_par%n_my,:,:)= &
           wave_block(i_t0,i_ztau_rec, &
             i_x_0_rec:i_x_0_rec+data_par%n_mx, &
             i_y_0_rec:i_y_0_rec+data_par%n_my,:,:)
     else
       data_io(1:data_par%n_mx,1,:,:)= &
           wave_block(i_t0,i_ztau_rec, &
             i_x_0_rec:i_x_0_rec+data_par%n_mx, &
             1,:,:)
     end if
   else if_comp_domain_data_1
     if(data_par%n_off_y > 1) then
       data_io(1:data_par%n_off_x,1:data_par%n_off_y,1,1)= &
           wave_block(i_t0,i_ztau_rec, &
             i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
             i_y_0_rec:i_y_0_rec+data_par%n_off_y,1,1)
     else
       data_io(1:data_par%n_off_x,1,1,1)= &
           wave_block(i_t0,i_ztau_rec, &
             i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
             1,1,1)
     end if
   end if if_comp_domain_data_1

   if(srite(trim(tag_data),data_io,n_bytes_io) .ne. n_bytes_io ) then
     write(0,*)'tag_data=',trim(tag_data)
     write(0,*)'i_t_data=',i_t_data
     call seperr('Problems with writing tag_data file')
   end if


  return	
  end subroutine output_data


  !-----------------------------------------------------------------------------
  subroutine input_data(tag_data,gen_par,data_par,wave_par,rev_extrap_par, &
                         force_block,wave_block)

   character(len=128) :: tag_data
   type (genrev3d_param_type) gen_par
   type (data_param_type) data_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:)   :: force_block
   real, dimension (:,:,:,:,:,:)   :: wave_block

   integer :: len_tag,i_t0,i_t_data,ret_sseek,nsseek_blocks
!   real   :: wave(wave_par%n_z_pad,wave_par%n_mx_pad,wave_par%n_my_pad)

   len_tag=len_trim(tag_data)

   i_t0=wave_par%i_t_step(1)
   i_t_data=nint((rev_extrap_par%t+rev_extrap_par%d_t-data_par%t_0)/data_par%d_t)+1
   write(0,*)'i_t_data=',i_t_data

   if(i_t_data < 1) call seperr("Asking for data before origin")
   if(i_t_data > data_par%n_t ) call seperr("Asking for data after end")


   nsseek_blocks=(i_t_data-1) +  &
                 ((data_par%i_sx-1)+(data_par%i_sy-1)*data_par%n_sx)*data_par%n_t

   ret_sseek=sseek_block(trim(tag_data),nsseek_blocks,n_bytes_io,0) 
   if (ret_sseek .ne. nsseek_blocks) then
   !{
     write(0,*)'ret_sseek=',ret_sseek
     write(0,*)'tag_data=',trim(tag_data)
     call seperr("Error block seeking data")
   !}
   end if

   if(sreed(trim(tag_data),data_io,n_bytes_io) .ne. n_bytes_io ) then
     write(0,*)'tag_data=',trim(tag_data)
     write(0,*)'i_t_data=',i_t_data
     call seperr('Problems with reading tag_data file')
   end if

   if_comp_domain_data_1 : if (gen_par%comp_domain(1:4) == 'midp') then

     force_block=0.

     if(data_par%n_my > 1) then

       force_block(&
          i_x_0_rec:i_x_0_rec+data_par%n_mx, &
          i_y_0_rec:i_y_0_rec+data_par%n_my,:,:)=&
            data_io(1:data_par%n_mx,1:data_par%n_my,:,:)

!       wave_block(i_t0,i_ztau_rec, &
!          i_x_0_rec:i_x_0_rec+data_par%n_mx, &
!          i_y_0_rec:i_y_0_rec+data_par%n_my,:,:)=&
!            data_io(1:data_par%n_mx,1:data_par%n_my,:,:)

     else
       force_block(&
          i_x_0_rec:i_x_0_rec+data_par%n_mx, &
          1,:,:)=&
            data_io(1:data_par%n_mx,1,:,:)

!       wave_block(i_t0,i_ztau_rec, &
!          i_x_0_rec:i_x_0_rec+data_par%n_mx, &
!          1,:,:)=&
!             data_io(1:data_par%n_mx,1,:,:)
     end if
   else if_comp_domain_data_1
     if(data_par%n_off_y > 1) then
       wave_block(i_t0,i_ztau_rec, &
          i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
          i_y_0_rec:i_y_0_rec+data_par%n_off_y,1,1)=&
            data_io(1:data_par%n_off_x,1:data_par%n_off_y,1,1)
     else
!       wave_block(i_t0,i_ztau_rec, &
!          i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
!          1,1,1)=&
!            data_io(1:data_par%n_off_x,1,1,1)
       wave_block(i_t0,i_ztau_rec, &
          i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
          1,1,1)=&
            data_io(1:data_par%n_off_x,1,1,1) 
!       wave_block(i_t0,i_ztau_rec, &
!          i_x_0_rec:i_x_0_rec+data_par%n_off_x, &
!          1,1,1)
     end if
   end if if_comp_domain_data_1


!   wave_block(i_t0,wave_par%n_z_pad/2,:,:,:,:)= wave_block(i_t0,wave_par%n_z_pad/2,:,:,:,:) + force_block
!   wave_block(i_t0,1,:,:,:,:)= wave_block(i_t0,1,:,:,:,:) + force_block
!   wave_block(i_t0,1,:,:,:,:)= force_block

   return	
   end subroutine input_data

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_data()
    integer ierr

    if(allocated(data_io)) then
       deallocate (data_io,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating data_io')
    end if

    return
  end subroutine clean_io_data


end module data_mod
