module genrev3d_types

  integer, parameter ::  status_init=0
  integer, parameter ::  mem_fixed_init=0
  integer, parameter ::  mem_n_w_inp_init=0
  integer, parameter ::  mem_n_mx_inp_init=0
  integer, parameter ::  mem_n_my_inp_init=0
  integer, parameter ::  mem_n_hx_inp_init=0
  integer, parameter ::  mem_n_hy_inp_init=0

  type genrev3d_info_type
     integer :: status
     integer :: mem_fixed
     integer :: mem_n_w_inp
     integer :: mem_n_mx_inp,mem_my_inp
     integer :: mem_n_hx_inp,mem_hy_inp
  end type  genrev3d_info_type

  type(genrev3d_info_type), parameter :: info_default = genrev3d_info_type &
  (status_init, &
  mem_fixed_init, &
  mem_n_w_inp_init, &
  mem_n_mx_inp_init, &
  mem_n_my_inp_init, &
  mem_n_hx_inp_init, &
  mem_n_hy_inp_init )

  logical, parameter :: verb_init=.true.
  logical, parameter :: debug1_init=.false.
  logical, parameter :: debug2_init=.false.
  logical, parameter :: timers_on_init=.false.
  logical, parameter :: image_zero_offset_init=.true.
  logical, parameter :: surf_offset_cig_init=.false.
  logical, parameter :: first_adj_init=.false.
  logical, parameter :: first_add_init=.false.
  logical, parameter :: check_stability_init=.false.
  logical, parameter :: tau_domain_init=.false.
  logical, parameter :: layered_mig_vel_init=.false.
  logical, parameter :: skip_source_prop_init=.false.

  character(len=128), parameter :: comp_domain_init='midpoint-offset'
  character(len=128), parameter :: oper_type_init='const_slow'
  character(len=128), parameter :: domain_range_init='data_image'

  integer, parameter :: n_threads_init=-1
  integer, parameter :: max_memory_init=134217728
  integer, parameter :: non_data_memory_init=0

  integer, parameter :: n_data_dim_init=4

  real, parameter :: slow_const_init=-1.

  real, parameter :: sou_depth_init=0.
  real, parameter :: rec_depth_init=0.

  integer, parameter :: integer_init=huge(1)/10.
  real, parameter :: real_init=huge(1.)/10.

  type genrev3d_param_type
     logical :: verb
     logical :: debug1
     logical :: debug2
     logical :: timers_on
     logical :: image_zero_offset
     logical :: surf_offset_cig
     logical :: first_adj
     logical :: first_add
     logical :: check_stability
     logical :: tau_domain
     logical :: layered_mig_vel
     logical :: skip_source_prop

! comp_domain - computational domain (shot, midpoint-offset)
! oper_type  - type of propgation operator(const_slow, full_slow)
! domain_range  - domain and range of linear operator (data_image)
     character(len=128) :: comp_domain
     character(len=128) :: oper_type
     character(len=128) :: domain_range

! n_threads  - number of threads
! max_memory  - maximum amount of memory that can be used in total
! non_data_memory  - maximum amount of memory that can be used not for the data
     integer :: n_threads
     integer :: max_memory
     integer :: non_data_memory

     integer :: n_data_dim

! slow_const  - constant slowness (if oper_type=const_slow)
! sou_depth  - depth of sources
! rec_depth  - depth of receivers
     real :: slow_const
     real :: sou_depth
     real :: rec_depth

  end type  genrev3d_param_type
end module genrev3d_types
