module image_mod

  use genrev3d_types
  use image_types
  use slow_types
  use wave_types
  use rev_extrap_types


  implicit none

  integer, external ::  sreed,srite,sseek_block
  logical, private :: adj,add,image_zero_offset
  integer, private :: n_bytes_io_image,n_bytes_io_wave,i_hx_0,i_hy_0
  integer, private :: n_ztau_image,n_ztau_wave
  integer, private :: i_hx_0_image,i_ztau_0_image

  logical, private :: image_memory ! Guojian
  logical, private :: write_image_hz  ! Guojian

  real, allocatable, dimension (:,:), private :: image_io,image_io_hz         !Guojian
  real, allocatable, dimension (:,:), private :: wave_io
  real, allocatable, dimension (:,:,:,:), private :: image_shot,image_shot_hz !Guojian
  
contains

  !-----------------------------------------------------------------------------
  subroutine init_io_image(gen_par,rev_extrap_par,wave_par,image_par)

  implicit none

    type (genrev3d_param_type) gen_par
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par
    type (image_param_type) image_par

    integer ierr

    !+++
 

    adj=rev_extrap_par%adj
    add=rev_extrap_par%add
    write_image_hz=image_par%write_image_hz
    write(0,*)'write_image_hz=',write_image_hz
    image_memory=image_par%image_memory
    write(0,*)'image_memory=',image_memory

    if_comp_domain_image_1 : if (gen_par%comp_domain(1:4) == 'midp') then
      if(gen_par%tau_domain) then
        n_ztau_image=image_par%n_tau
        i_ztau_0_image=image_par%tau_0/wave_par%d_tau + 1
    write(0,*)'i_ztau_0_image=',i_ztau_0_image
        if(gen_par%image_zero_offset) then
          image_zero_offset=.true.
          n_bytes_io_image=n_ztau_image*image_par%n_mx*image_par%esize
    write(0,*)'image_par%n_ztau_image',n_ztau_image
    write(0,*)'image_par%n_mx',image_par%n_mx
    write(0,*)'image_par%n_my',image_par%n_my
    write(0,*)'n_bytes_io_image=',n_bytes_io_image

          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          if (write_image_hz) then
            allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
          end if

          i_hx_0=(-wave_par%hx_0/wave_par%d_hx)+1
    write(0,*)'i_hx_0=',i_hx_0
          if(i_hx_0 < 1)  &
             call seperr("Trying to image at zero offset but i_hx_0 < 1") 
          if(i_hx_0 > wave_par%n_hx)  &
             call seperr("Trying to image at zero offset but i_hx_0 > n_hx") 
          i_hy_0=(-wave_par%hy_0/wave_par%d_hy)+1
          if(i_hy_0 < 1)  &
             call seperr("Trying to image at zero offset but i_hy_0 < 1") 
          if(i_hy_0 > wave_par%n_hy)  &
             call seperr("Trying to image at zero offset but i_hy_0 > n_hy") 
        else
          image_zero_offset=.false.
          n_bytes_io_image=n_ztau_image*image_par%n_mx*image_par%esize
    write(0,*)'image_par%n_ztau_image',n_ztau_image
    write(0,*)'image_par%n_mx',image_par%n_mx
    write(0,*)'image_par%n_my',image_par%n_my
    write(0,*)'n_bytes_io_image=',n_bytes_io_image
          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if (write_image_hz) then
            allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
          end if
        end if
      else
        n_ztau_image=image_par%n_z
        i_ztau_0_image=image_par%z_0/wave_par%d_z + 1
    write(0,*)'i_ztau_0_image=',i_ztau_0_image
        if(gen_par%image_zero_offset) then
          image_zero_offset=.true.
          n_bytes_io_image=image_par%n_z*image_par%n_mx*image_par%esize
    write(0,*)'image_par%n_ztau_image',n_ztau_image
    write(0,*)'image_par%n_mx',image_par%n_mx
    write(0,*)'image_par%n_my',image_par%n_my
    write(0,*)'n_bytes_io_image=',n_bytes_io_image

          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          if (write_image_hz) then
            allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
          end if
          i_hx_0=(-wave_par%hx_0/wave_par%d_hx)+1
    write(0,*)'i_hx_0=',i_hx_0
          if(i_hx_0 < 1)  &
             call seperr("Trying to image at zero offset but i_hx_0 < 1") 
          if(i_hx_0 > wave_par%n_hx)  &
             call seperr("Trying to image at zero offset but i_hx_0 > n_hx") 
          i_hy_0=(-wave_par%hy_0/wave_par%d_hy)+1
          if(i_hy_0 < 1)  &
             call seperr("Trying to image at zero offset but i_hy_0 < 1") 
          if(i_hy_0 > wave_par%n_hy)  &
             call seperr("Trying to image at zero offset but i_hy_0 > n_hy") 
        else
          image_zero_offset=.false.
          n_bytes_io_image=image_par%n_z*image_par%n_mx*image_par%esize
    write(0,*)'image_par%n_ztau_image',n_ztau_image
    write(0,*)'image_par%n_mx',image_par%n_mx
    write(0,*)'image_par%n_my',image_par%n_my
    write(0,*)'n_bytes_io_image=',n_bytes_io_image
          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if (write_image_hz) then
            allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
          end if
        end if
      end if
    else if_comp_domain_image_1


      if(gen_par%tau_domain) then
      !{
        n_ztau_image=image_par%n_tau
        n_ztau_wave=wave_par%n_tau
        i_ztau_0_image=image_par%tau_0/wave_par%d_tau + 1
    write(0,*)'i_ztau_0_image=',i_ztau_0_image

        if(gen_par%image_zero_offset) then
          image_zero_offset=.true.
          n_bytes_io_image=n_ztau_image*image_par%n_mx*image_par%esize
          n_bytes_io_wave=n_ztau_wave*wave_par%n_mx*wave_par%esize
          i_hx_0_image=0

          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          if (write_image_hz) then
          allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
          end if

          allocate (wave_io(n_ztau_wave,wave_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
          if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
          wave_io=0.
        else
          image_zero_offset=.false.
          n_bytes_io_image=n_ztau_image*image_par%n_mx*image_par%esize
          n_bytes_io_wave=n_ztau_wave*wave_par%n_mx*wave_par%esize
!          i_hx_0_image=image_par%n_hx/2
          i_hx_0_image=(image_par%hx_0/image_par%d_hx)
          i_hx_0_image=nint(-image_par%hx_0/image_par%d_hx)+1
      write(0,*)'(Image) i_hx_0_image=',i_hx_0_image

          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          if (write_image_hz) then
          allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
          end if


          allocate (wave_io(n_ztau_wave,wave_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
          if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
          wave_io=0.
        end if
      !}
      else
      !{
        n_ztau_image=image_par%n_z
        n_ztau_wave=wave_par%n_z
        i_ztau_0_image=image_par%z_0/wave_par%d_z + 1
    write(0,*)'i_ztau_0_image=',i_ztau_0_image

        if(gen_par%image_zero_offset) then
        !{
          image_zero_offset=.true.
          n_bytes_io_image=image_par%n_z*image_par%n_mx*image_par%esize
          n_bytes_io_wave=wave_par%n_z*wave_par%n_mx*wave_par%esize
          i_hx_0_image=0

    write(0,*)'n_bytes_io_image=',n_bytes_io_image
    write(0,*)'n_bytes_io_wave=',n_bytes_io_wave

!  Guojian
          if(image_memory) then
            allocate (image_shot(n_ztau_image,image_par%n_mx,image_par%n_hx,image_par%n_t_lag), stat=ierr)
            if(gen_par%debug2) write(0,*)'Allocate image_shot with shape=',shape(image_shot)
            if(ierr .ne. 0) call seperr("Problems with allocation of image_shot")
            image_shot=0.
          end if
! end Guojian


          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
          allocate (wave_io(n_ztau_wave,wave_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
          if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
          wave_io=0.
        !}
        else
        !{
          image_zero_offset=.false.
          n_bytes_io_image=image_par%n_z*image_par%n_mx*image_par%esize
          n_bytes_io_wave=wave_par%n_z*wave_par%n_mx*wave_par%esize
!          i_hx_0_image=image_par%n_hx/2
          i_hx_0_image=nint(-image_par%hx_0/image_par%d_hx)+1
          write(0,*)'(Image) i_hx_0_image=',i_hx_0_image

!  Guojian
          if(image_memory) then
            allocate (image_shot(n_ztau_image,image_par%n_mx,image_par%n_hx,image_par%n_t_lag), stat=ierr)
            if(gen_par%debug2) write(0,*)'Allocate image_shot with shape=',shape(image_shot)
            if(ierr .ne. 0) call seperr("Problems with allocation of image_shot")
            image_shot=0.

            if (write_image_hz) then
              allocate (image_shot_hz(n_ztau_image,image_par%n_mx,image_par%n_hx,image_par%n_t_lag), stat=ierr)
              if(gen_par%debug2) write(0,*)'Allocate image_shot_hz with shape=',shape(image_shot_hz)
              if(ierr .ne. 0) call seperr("Problems with allocation of image_shot_hz")
              image_shot_hz=0.
            end if

          end if
! end Guojian

          allocate (image_io(n_ztau_image,image_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
          if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
!  Guojian
          if (write_image_hz) then
            allocate (image_io_hz(n_ztau_image,image_par%n_mx), stat=ierr)  
            if(ierr .ne. 0) call seperr("Problems with allocation of image_io_hz")
            if(gen_par%debug2) write(0,*)'Allocate image_io_hz with shape=',shape(image_io_hz)
          end if
!  end Guojian



          allocate (wave_io(n_ztau_wave,wave_par%n_mx), stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
          if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
          wave_io=0.
        !}
        end if
      !}
      end if
    end if if_comp_domain_image_1
  
  end subroutine init_io_image


  !-----------------------------------------------------------------------------
  subroutine output_image(tag_image,tag_image_hz,gen_par,image_par, &
                          wave_par,rev_extrap_par, &
                          max_it_snap_image, &
                          wave_block)
! record to memory.
   implicit none

   character(len=128) :: tag_image,tag_image_hz ! 'tag_image_hz' added by Guojian
   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:,:,:)   :: wave_block
   integer :: max_it_snap_image !Guojian

   character(len=128) :: tag_wave
   logical sum_image_file,write_image
   integer :: len_tag_image,len_tag_image_hz,len_tag_wave !len_tag_image_hz added by Guojian
   integer :: i_ztau,i_hx,i_hy,i_my,i_mx
   integer :: i_t_image,i_t_wave
   integer :: i_mx_0_image,i_my_0_image,i_mx_wave_beg,i_mx_wave_end
   integer :: min_it_snap_image
   integer :: i_t_lag_beg,i_t_lag_end,i_t_lag
   integer :: i_mx_image,i_mx_wave
   integer :: i_hx_shift_sou,i_hx_shift_rec
   integer  nsseek_blocks,ret_sseek,block_wave,ret_sseek_ok
   integer tmpi_t_lag,wavepar_itstep1,n_threads,i_thread  ! for parallelization


   n_threads=rev_extrap_par%n_threads
   len_tag_image=len_trim(tag_image)
   len_tag_image_hz=len_trim(tag_image_hz)  ! Guojian
   tag_wave=trim(wave_par%tag_2)
   len_tag_wave=len_trim(tag_wave)

   if_comp_domain_image_2 : if (gen_par%comp_domain(1:4) == 'midp') then
     if(add) then
     !{
      ! add to the image
      !          write(0,*)'add_to_image seeek in image'
      if(gen_par%debug2) write(0,*)'Adding to ',tag_image(1:len_tag_image)
       call seperr('add to image Not implemented yet')
     !} 
     else 
     !{
       if(gen_par%debug2) write(0,*)'Writing to ',tag_image(1:len_tag_image)
       if(image_zero_offset) then
       !{
         do i_my=1,image_par%n_my
           do i_mx=1,image_par%n_mx
             do i_ztau=1,n_ztau_image
               image_io(i_ztau,i_mx)= &
                  wave_block(wave_par%i_t_step(1),i_ztau+i_ztau_0_image,i_mx,i_my,i_hx_0,i_hy_0)
             end do
           end do
           if(srite(trim(tag_image),image_io,n_bytes_io_image) /=  &
              n_bytes_io_image ) then
             write(0,*)'tag_image=',trim(tag_image)
             call seperr('Problems with writing tag_image file')
           end if
         end do
       !}
       else
       !{
                 write(0,*)'i_t_step(1)=',wave_par%i_t_step(1)
                 write(0,*)'n_ztau_image=',n_ztau_image
                 write(0,*)'n_mx=',image_par%n_mx
                 write(0,*)'n_hx=',image_par%n_hx
                 write(0,*)'n_my=',image_par%n_my
                 write(0,*)'n_hy=',image_par%n_hy
         do i_hy=1,image_par%n_hy
           do i_hx=1,image_par%n_hx
             do i_my=1,image_par%n_my
                do i_mx=1,image_par%n_mx
                  do i_ztau=1,n_ztau_image
                     image_io(i_ztau,i_mx)= &
                     wave_block(wave_par%i_t_step(1),i_ztau+i_ztau_0_image,i_mx,i_my,i_hx,i_hy)
                  end do
                  if (write_image_hz) then
                    do i_ztau=1,n_ztau_image
                       image_io_hz(i_ztau,i_mx)= &
                       wave_block(wave_par%i_t_step(1),i_ztau+i_ztau_0_image,i_mx,i_my,i_hx,i_hy)
                    end do
                  end if

                end do
                if(srite(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
                 write(0,*)'tag_image=',trim(tag_image)
                 write(0,*)'i_my=',i_my
                 write(0,*)'i_hx=',i_hx
                 write(0,*)'i_hy=',i_hy
                 call seperr('Problems with writing tag_image file')
                end if
             end do
           end do
         end do

                 write(0,*)'Done with image'
    
       !}
       end if
     !} 
     end if
   else if_comp_domain_image_2

!     i_t_lag_beg=max(-i_t_wave+1,nint(image_par%t_0_lag/image_par%d_t_lag))
!     i_t_lag_end=min(image_par%n_t-i_t_wave,i_t_lag_beg+image_par%n_t_lag-1)
     if(image_par%n_t_lag /= 1) then
       i_t_lag_beg=nint(image_par%t_0_lag/image_par%d_t_lag)
       i_t_lag_end=i_t_lag_beg+image_par%n_t_lag-1
     else
       i_t_lag_beg=1
       i_t_lag_end=1
     end if

     if (image_memory) then
!    Guojian
        write_image = (rev_extrap_par%i_t_extrap == max_it_snap_image)
     write(0,*)'write_image=',write_image
        sum_image_file = write_image .and. &
          (((rev_extrap_par%i_sx /= 1) .or. (rev_extrap_par%i_sy /= 1)) .and. &
           (.not. gen_par%surf_offset_cig) &
          )
!    end Guojian
     else
          min_it_snap_image=i_t_lag_end
          sum_image_file = &
          (((rev_extrap_par%i_t_extrap /= min_it_snap_image) .or.  &
            ((rev_extrap_par%i_sx /= 1) .or. (rev_extrap_par%i_sy /= 1))) &
          )
     end if
     write(0,*)'rev_extrap_par%i_t_extrap =',rev_extrap_par%i_t_extrap 
     write(0,*)'sum_image_file=',sum_image_file


     i_t_wave=nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+1
!     i_t_wave=nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+3

     i_t_image= image_par%n_t - &
                i_t_lag_end - &
                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))-2

!     i_t_image= image_par%n_t + &
!                i_t_lag_beg - &
!                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))-2

!                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+1)

     i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1
     i_my_0_image=nint((wave_par%my_0-image_par%my_0)/image_par%d_my)+1

     i_mx_wave_beg=max(i_hx_0_image+1,-i_mx_0_image+2)
     i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image), &
                       image_par%n_mx-i_mx_0_image+1)


!     if(gen_par%debug2) write(0,*)'i_t_wave=',i_t_wave
!     if(gen_par%debug2) write(0,*)'i_t_image=',i_t_image
!     if(gen_par%debug2) write(0,*)'i_mx_0_image=',i_mx_0_image
!     if(gen_par%debug2) write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!     if(gen_par%debug2) write(0,*)'i_mx_wave_end=',i_mx_wave_end
!     if(gen_par%debug2) write(0,*)'i_my_0_image=',i_my_0_image
!     if(gen_par%debug2) write(0,*)'i_t_lag_beg=',i_t_lag_beg
!     if(gen_par%debug2) write(0,*)'i_t_lag_end=',i_t_lag_end



     if(sum_image_file)  then
     !{
       if(image_par%write_image_progress) then
         nsseek_blocks= -image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                         image_par%n_hy
!         ret_sseek_ok= (i_t_image-2)* &
!         ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
         ret_sseek_ok= ( &
                                 (i_t_image-2)*image_par%n_t_lag)* &
                                 image_par%n_my*image_par%n_hx *image_par%n_hy
       else
         nsseek_blocks= -image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                         image_par%n_hy
         ret_sseek_ok=0
!         ret_sseek_ok=i_t_lag - i_t_lag_beg
       end if

!       write(0,*)'First seek on Image: nsseek_blocks=',nsseek_blocks

       ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_io_image,1) 
       if (ret_sseek .ne. ret_sseek_ok) then
       !{
         write(0,*)'nsseek_blocks=',nsseek_blocks
         write(0,*)'ret_sseek=',ret_sseek
         write(0,*)'ret_sseek_ok=',ret_sseek_ok
         write(0,*)'tag_image=',trim(tag_image)
         call seperr("Error block seeking image")
       !}
       end if
       if (write_image_hz) then
         ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_io_image,1) 
         if (ret_sseek .ne. ret_sseek_ok) then
         !{
           write(0,*)'nsseek_blocks=',nsseek_blocks
           write(0,*)'ret_sseek=',ret_sseek
           write(0,*)'ret_sseek_ok=',ret_sseek_ok
           write(0,*)'tag_image_hz=',trim(tag_image_hz)
           call seperr("Error block seeking image")
         !}
         end if
       end if

     !}
     end if

     do i_t_lag=i_t_lag_beg,i_t_lag_end
     !{

     if(gen_par%debug2) write(0,*)'i_t_lag=',i_t_lag

     do i_hy=1,image_par%n_hy
     !{
       do i_my=1,image_par%n_my
       !{

         block_wave= (i_t_wave-1+i_t_lag)*wave_par%n_my
         nsseek_blocks= block_wave + (i_my-1)*wave_par%n_hx
!     write(0,*)'block_wave=',block_wave
!         write(0,*)'nsseek_blocks (wave)=',nsseek_blocks

         ret_sseek=sseek_block(trim(tag_wave),nsseek_blocks,n_bytes_io_wave,0) 
         if (ret_sseek .ne. nsseek_blocks) then
         !{
           write(0,*)'nsseek_blocks=',nsseek_blocks
           write(0,*)'ret_sseek=',ret_sseek
           write(0,*)'tag_wave=',trim(tag_wave)
           call seperr("Error block seeking wave")
         !}
         end if
         if(sreed(trim(tag_wave),wave_io,n_bytes_io_wave)/=n_bytes_io_wave) then
           write(0,*)'tag_wave=',trim(tag_wave)
           call seperr('Problems with reading tag_wave file')
         end if


         do i_hx=1,image_par%n_hx
         !{
           if(image_zero_offset) then
             i_hx_shift_sou=0
             i_hx_shift_rec=0
           else
             i_hx_shift_sou=i_hx_0_image-i_hx
             i_hx_shift_rec=-(i_hx_0_image-i_hx)
           end if
           
           if(sum_image_file) then
           !{

             if(sreed(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
               write(0,*)'tag_image=',trim(tag_image)
               call seperr('Problems with reading tag_image file')
             end if
             if (write_image_hz) then
               if(sreed(trim(tag_image_hz),image_io_hz,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
                 write(0,*)'tag_image=',trim(tag_image_hz)
                 call seperr('Problems with reading tag_image_hz file')
               end if
             end if
             if (image_memory) then
               image_shot(:,:,i_hx,i_t_lag-i_t_lag_beg+1)= &
                  image_shot(:,:,i_hx,i_t_lag-i_t_lag_beg+1)+image_io(:,:)
               if (write_image_hz) then
               image_shot_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)= &
                  image_shot_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)+image_io_hz(:,:)
               end if
             end if

             if(image_par%write_image_progress) then
               nsseek_blocks= image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                         image_par%n_hy -1
!               ret_sseek_ok= (i_t_image-1)* &
               ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
                              (i_t_image-1)*image_par%n_t_lag)* &
                              image_par%n_my*image_par%n_hx *image_par%n_hy + &
                              (i_hx-1)* &
                              image_par%n_my*image_par%n_hy 
             else
               nsseek_blocks= -1
               ret_sseek_ok= (i_t_lag -i_t_lag_beg)* &
                              image_par%n_my*image_par%n_hx*image_par%n_hy + &
                             (i_hx-1)* &
                              image_par%n_my*image_par%n_hy 
             end if

!       write(0,*)'in hx loop seek on Image after read: nsseek_blocks=',nsseek_blocks

             ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_io_image,1) 
             if (ret_sseek .ne. ret_sseek_ok) then
             !{
               write(0,*)'nsseek_blocks=',nsseek_blocks
               write(0,*)'ret_sseek=',ret_sseek
               write(0,*)'ret_sseek_ok=',ret_sseek_ok
               write(0,*)'tag_image=',trim(tag_image)
               call seperr("Error block seeking image")
             !}
             end if
!               write(0,*)'ret_sseek=',ret_sseek
             if (write_image_hz) then
               ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_io_image,1) 
               if (ret_sseek .ne. ret_sseek_ok) then
               !{
                 write(0,*)'nsseek_blocks=',nsseek_blocks
                 write(0,*)'ret_sseek=',ret_sseek
                 write(0,*)'ret_sseek_ok=',ret_sseek_ok
                 write(0,*)'tag_image_hz=',trim(tag_image_hz)
                 call seperr("Error block seeking image")
               !}
               end if
             end if


           !}
           else
           !{
             if (.not. image_memory) then
             image_io=0.
             if (image_par%write_image_hz) then
               image_io_hz=0.
             end if
             end if
           !}
           end if
          
           if (image_memory) then 
           !{
             tmpi_t_lag=i_t_lag-i_t_lag_beg+1
             wavepar_itstep1=wave_par%i_t_step(1)
!$omp parallel default (private), shared (image_shot,image_shot_hz,wave_io,wave_block), &
!$omp& shared(n_threads,i_mx_wave_beg,i_mx_wave_end,i_mx_0_image,n_ztau_image), &
!$omp& shared(i_hx,tmpi_t_lag,i_hx_shift_sou,wavepar_itstep1,i_hx_shift_rec,i_my)
!$omp do

             do i_thread=1,n_threads
               do i_mx_wave=i_mx_wave_beg-1+i_thread,i_mx_wave_end,n_threads
                 i_mx_image=i_mx_wave+i_mx_0_image-1 
                 do i_ztau=1,n_ztau_image
                    image_shot(i_ztau,i_mx_image,i_hx,tmpi_t_lag)=&
                    image_shot(i_ztau,i_mx_image,i_hx,tmpi_t_lag)+&
                    wave_io(i_ztau+i_ztau_0_image,i_mx_wave+i_hx_shift_sou)* &
                    wave_block(wavepar_itstep1, &
                       i_ztau+i_ztau_0_image,i_mx_wave+i_hx_shift_rec,i_my,1,1)      
                 end do
               end do
             end do

             if (write_image_hz) then
               do i_thread=1,n_threads
                 do i_mx_wave=i_mx_wave_beg-1+i_thread,i_mx_wave_end,n_threads
                   i_mx_image=i_mx_wave+i_mx_0_image-1 
                   do i_ztau=1,n_ztau_image
                      if (i_ztau+i_ztau_0_image+i_hx_shift_sou >0 .and.i_ztau+i_ztau_0_image+i_hx_shift_sou <=n_ztau_wave) then
                        if ( i_ztau+i_ztau_0_image+i_hx_shift_rec >0 .and.i_ztau+i_ztau_0_image+i_hx_shift_rec <=n_ztau_wave ) then 
                           image_shot_hz(i_ztau,i_mx_image,i_hx,tmpi_t_lag)=&
                           image_shot_hz(i_ztau,i_mx_image,i_hx,tmpi_t_lag)+&
                           wave_io(i_ztau+i_ztau_0_image+i_hx_shift_sou,i_mx_wave)* &
                           wave_block(wavepar_itstep1,i_ztau+i_ztau_0_image+i_hx_shift_rec,i_mx_wave,i_my,1,1)      
                        end if
                      end if
                   end do
                 end do
               end do
              end if


!             do i_thread=1,n_threads
!               do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
!                 i_mx_image=i_mx_wave+i_mx_0_image-1 
!                 do i_ztau=1,n_ztau_image
!                    image_shot(i_ztau,i_mx_image,i_hx,i_t_lag-i_t_lag_beg+1)=&
!                    image_shot(i_ztau,i_mx_image,i_hx,i_t_lag-i_t_lag_beg+1)+&
!                    wave_io(i_ztau,i_mx_wave+i_hx_shift_sou)* &
!                    wave_block(wave_par%i_t_step(1), &
!                       i_ztau,i_mx_wave+i_hx_shift_rec,i_my,1,1)      
!                 end do
!               end do
!             end do

!$omp end parallel

           !}
           else
           !{
             do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
               i_mx_image=i_mx_wave+i_mx_0_image-1
               do i_ztau=1,n_ztau_image
                  image_io(i_ztau,i_mx_image)=image_io(i_ztau,i_mx_image) + &
                  wave_io(i_ztau+i_ztau_0_image,i_mx_wave+i_hx_shift_sou)* &
                  wave_block(wave_par%i_t_step(1), &
                       i_ztau+i_ztau_0_image,i_mx_wave+i_hx_shift_rec,i_my,1,1)


               end do
             end do
           !}
           end if

           if (image_memory ) then
           !{
             if(write_image) then
                image_io(:,:) =image_shot(:,:,i_hx,i_t_lag-i_t_lag_beg+1)
                image_shot(:,:,i_hx,i_t_lag-i_t_lag_beg+1)=0.
                if (write_image_hz) then
                  image_io_hz(:,:) =image_shot_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)
                  image_shot_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)=0.
                end if
                if(srite(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
                  write(0,*)'tag_image=',trim(tag_image)
                  call seperr('Problems with writing tag_image file')
                end if
                if (write_image_hz) then
                  if(srite(trim(tag_image_hz),image_io_hz,n_bytes_io_image) /= &
                    n_bytes_io_image ) then
                    write(0,*)'tag_image_hz=',trim(tag_image_hz)
                    call seperr('Problems with writing tag_image_hz file')
                  end if
                end if
             end if 
           !}
           else
           !{
             if(srite(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
               write(0,*)'tag_image=',trim(tag_image)
               call seperr('Problems with writing tag_image file')
             end if
           !}
           end if
 
           if(sum_image_file .and. (i_hx /= image_par%n_hx)) then
           !{
             if(image_par%write_image_progress) then
             !{
               nsseek_blocks=-image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                         image_par%n_hy
!               ret_sseek_ok= (i_t_image-2)* &
               ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
                              (i_t_image-2)*image_par%n_t_lag)* &
                              image_par%n_my*image_par%n_hx *image_par%n_hy + &
                              (i_hx)* &
                              image_par%n_my*image_par%n_hy 
             !}
             else
             !{
               nsseek_blocks= 0
               ret_sseek_ok= (i_t_lag -i_t_lag_beg)* &
                              image_par%n_my*image_par%n_hx*image_par%n_hy + &
                             (i_hx)* &
                              image_par%n_my*image_par%n_hy 
             !}
             end if

             ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_io_image,1) 
             if (ret_sseek .ne. ret_sseek_ok) then
             !{
               write(0,*)'nsseek_blocks=',nsseek_blocks
               write(0,*)'ret_sseek=',ret_sseek
               write(0,*)'ret_sseek_ok=',ret_sseek_ok
               write(0,*)'tag_image=',trim(tag_image)
               call seperr("Error block seeking image")
             !}
             end if
             if (write_image_hz) then
             !{
               ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_io_image,1) 
               if (ret_sseek .ne. ret_sseek_ok) then
               !{
                 write(0,*)'nsseek_blocks=',nsseek_blocks
                 write(0,*)'ret_sseek=',ret_sseek
                 write(0,*)'ret_sseek_ok=',ret_sseek_ok
                 write(0,*)'tag_image_hz=',trim(tag_image_hz)
                 call seperr("Error block seeking image")
               !}
               end if
             !}
             end if
           !}
           end if

 
         !}
         end do
       !}
       end do
     !}
     end do

     !}
     end do


   end if if_comp_domain_image_2


   return	
  end subroutine output_image


  !-----------------------------------------------------------------------------
  !!!!!!!!!!!! I don't change anything for input_image for vertical offset :Guojian Shan!!!
  subroutine input_image(tag_image,gen_par,image_par,wave_par,wave_block)

   implicit none


   character(len=128) :: tag_image
   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (wave_param_type) :: wave_par
   real, dimension (:,:,:,:,:,:)   :: wave_block

   integer :: len_tag_image,i_hx,i_hy,i_my

   len_tag_image=len_trim(tag_image)

   if(add) then
   !{
    ! add to the image
    !          write(0,*)'add_to_image seeek in image'
    if(gen_par%debug2) write(0,*)'Adding to ',tag_image(1:len_tag_image)
     call seperr('add to image Not implemented yet')
   !} 
   else 
   !{
     if(gen_par%debug2) write(0,*)'Reading from ',tag_image(1:len_tag_image)
     if(image_zero_offset) then
     !{
       do i_my=1,image_par%n_my
         if(sreed(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
           write(0,*)'tag_image=',trim(tag_image)
           call seperr('Problems with reading tag_image file')
         end if
         wave_block(wave_par%i_t_step(1),:,:,i_my,i_hx_0,i_hy_0)=image_io
       end do
     !}
     else
     !{
       do i_hy=1,image_par%n_hy
         do i_hx=1,image_par%n_hx
           do i_my=1,image_par%n_my
             if(sreed(trim(tag_image),image_io,n_bytes_io_image) /= &
                  n_bytes_io_image ) then
               write(0,*)'tag_image=',trim(tag_image)
               write(0,*)'i_my=',i_my
               write(0,*)'i_hx=',i_hx
               write(0,*)'i_hy=',i_hy
               call seperr('Problems with reading tag_image file')
             end if

             wave_block(wave_par%i_t_step(1),:,:,i_my,i_hx,i_hy)=image_io
           end do
         end do
       end do

     !}
     end if
   !} 
   end if


    return	
  end subroutine input_image

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_image()

    integer ierr

    if(allocated(image_io)) then
       deallocate (image_io,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_io')
    end if
    if (write_image_hz) then
      if(allocated(image_io_hz)) then
       deallocate (image_io_hz,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_io_hz')
      end if
    end if
    if(allocated(image_shot)) then
       deallocate (image_shot,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_shot')
    end if
    if (write_image_hz) then
      if(allocated(image_shot_hz)) then
       deallocate (image_shot_hz,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_shot_hz')
      end if
    end if

    return
  end subroutine clean_io_image


end module image_mod
