!*$*OPTIMIZE(5)
!*$*ROUNDOFF(3)
!*$*SCALAROPTIMIZE(3)

module rev_extrap_gen_mod
  use genrev3d_types
  use data_types
  use image_types
  use slow_types
  use wave_types
  use wavelet_types
  use rev_extrap_types

  use wave_mod
  use wavelet_mod
  use image_mod
  use data_mod
  use slow_mod

  use timers_mod

!  use down_gen_ser_mod
!  use image_mod
!  use filter_data_mod

  implicit none
contains
  function rev_extrap_gen( &
    gen_par,data_par,image_par,rev_extrap_par,slow_par,wave_par,wavelet_par, &
    rev_extrap_oper,rev_scatter_oper, &
    slow_block,&
    force_block,&
    wave_block,&
    wavelet) &
  result(stat)

  interface 
    function rev_extrap_oper( &
                             rev_extrap_par, &
                             wave_par, &
                             force_block, &
                             wave_block, slow_block) &
                             result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use data_types, only: data_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:)                :: force_block
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real, optional,  dimension (:,:,:,:,:,:)  :: slow_block
    end function rev_extrap_oper
  
    function rev_scatter_oper(adj,add &
                   ) &
                   result(stat)
      integer                                   :: stat
      logical                                   :: adj,add
    end function rev_scatter_oper
  
  end interface

    integer stat
    type(genrev3d_param_type) :: gen_par
    type (data_param_type) :: data_par
    type (image_param_type) :: image_par
    type (rev_extrap_param_type) :: rev_extrap_par
    type (slow_param_type) :: slow_par
    type (wave_param_type) :: wave_par
    type (wavelet_param_type) :: wavelet_par

    optional rev_scatter_oper


    integer stat_par(gen_par%n_threads)
    real, dimension    (:,:,:,:,:,:) :: slow_block
    real, dimension (:,:,:,:) :: force_block
    real, dimension (:,:,:,:,:,:) :: wave_block
    real, dimension (:,:,:) :: wavelet

    integer, external :: omp_get_thread_num

    integer ierr,i_t_snap
    integer rat_t_snap_wave,rat_t_snap_image,rat_t_snap_data
    integer rat_t_snap_wave_sou,rat_t_snap_wave_tmp
    integer min_it_snap_wave,min_it_snap_image,min_it_snap_data
    integer min_it_snap_wave_sou,min_it_snap_wave_tmp
    integer max_it_snap_wave,max_it_snap_image,max_it_snap_data
    integer max_it_snap_wave_sou,max_it_snap_wave_tmp
    integer i_sx,i_sy,i_t_extrap
    real t_extrap,epsi_image_d_t,epsi_wave_d_t,epsi_data_d_t

    stat_par=0 ! in case that we never call the inner routines

    epsi_image_d_t=.01*image_par%d_t
    epsi_wave_d_t=.01*wave_par%d_t
    epsi_data_d_t=.01*data_par%d_t

    ! read appropriate wavelet files
    if(wavelet_par%file == WAVELET_INPUT) then
      if(gen_par%debug1) write(0,*)'Read Wavelet'
      call read_wavelet(wavelet_par%tag,wavelet_par,wavelet)
    end if

    if_comp_domain_1 : if (gen_par%comp_domain(1:4) == 'shot') then

      if_adj_1 : if (rev_extrap_par%adj) then

        if(gen_par%debug1) write(0,*)'image_par-t_0=',image_par%t_0
        if(gen_par%debug1) write(0,*)'rev_extrap_par-t_0_first=',rev_extrap_par%t_0_first
        min_it_snap_wave_tmp=nint((image_par%t_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)

        max_it_snap_wave_tmp=rev_extrap_par%n_t_first
        rat_t_snap_wave_tmp=nint(image_par%d_t/rev_extrap_par%d_t_first)
        if(rat_t_snap_wave_tmp == 0) rat_t_snap_wave_tmp=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_tmp=',min_it_snap_wave_tmp
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_tmp=',max_it_snap_wave_tmp
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_tmp=',rat_t_snap_wave_tmp

        min_it_snap_wave_sou=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)
        max_it_snap_wave_sou=rev_extrap_par%n_t_first
        rat_t_snap_wave_sou=-nint(wave_par%d_t/rev_extrap_par%d_t_first)
        if(rat_t_snap_wave_sou == 0) rat_t_snap_wave_sou=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_sou=',min_it_snap_wave_sou
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_sou=',max_it_snap_wave_sou
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_sou=',rat_t_snap_wave_sou

        min_it_snap_wave=0
        max_it_snap_wave=rev_extrap_par%n_t_second
        rat_t_snap_wave=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave == 0) rat_t_snap_wave=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave=',min_it_snap_wave
        if(gen_par%debug1) write(0,*)'max_it_snap_wave=',max_it_snap_wave
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave=',rat_t_snap_wave
  
!        min_it_snap_image=nint(image_par%t_0_lag/image_par%d_t_lag) + &
!                          image_par%n_t_lag-1 &
!                          +2
!        max_it_snap_image=rev_extrap_par%n_t_second+1+ &
!                            nint(image_par%t_0_lag/image_par%d_t_lag)

        min_it_snap_image=nint(image_par%t_0_lag/image_par%d_t_lag) + &
                          image_par%n_t_lag-1 
        max_it_snap_image=rev_extrap_par%n_t_second-1+ &
                            nint(image_par%t_0_lag/image_par%d_t_lag)


        rat_t_snap_image=-nint(image_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_image == 0) rat_t_snap_image=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
        if(gen_par%debug1) write(0,*)'max_it_snap_image=',max_it_snap_image
        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
        min_it_snap_data=0
        max_it_snap_data=data_par%n_t
        rat_t_snap_data=-nint(data_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_data == 0) rat_t_snap_data=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data

      else if_adj_1

        min_it_snap_wave=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_wave=rev_extrap_par%n_t_second
        rat_t_snap_wave=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave == 0) rat_t_snap_wave=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave=',min_it_snap_wave
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave=',rat_t_snap_wave
  
        min_it_snap_image=-1
        max_it_snap_image=-1
        rat_t_snap_image=-1

!        min_it_snap_image=nint((image_par%t_0-rev_extrap_par%t_0_second)/ &
!                              rev_extrap_par%d_t_second)
!        max_it_snap_image=rev_extrap_par%n_t_second
!        rat_t_snap_image=nint(image_par%d_t/rev_extrap_par%d_t_second)
!        if(rat_t_snap_image == 0) rat_t_snap_image=-1
!        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
!        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
        min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_data=rev_extrap_par%n_t_second-1
        rat_t_snap_data=nint(data_par%d_t/rev_extrap_par%d_t_second)

        if(rat_t_snap_data == 0) rat_t_snap_data=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data
      end if if_adj_1

    else if_comp_domain_1
      if_adj_2 : if (rev_extrap_par%adj) then

        min_it_snap_wave=0
        max_it_snap_wave=rev_extrap_par%n_t_second
        rat_t_snap_wave=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave == 0) rat_t_snap_wave=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave=',min_it_snap_wave
        if(gen_par%debug1) write(0,*)'max_it_snap_wave=',max_it_snap_wave
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave=',rat_t_snap_wave
  
        min_it_snap_image=rev_extrap_par%n_t_second - image_par%n_t_lag +1
        max_it_snap_image=rev_extrap_par%n_t_second

        rat_t_snap_image=1
        if(rat_t_snap_image == 0) rat_t_snap_image=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
        if(gen_par%debug1) write(0,*)'max_it_snap_image=',max_it_snap_image
        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image

        min_it_snap_data=0
        max_it_snap_data=data_par%n_t
        rat_t_snap_data=-nint(data_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_data == 0) rat_t_snap_data=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data

      else if_adj_2

        min_it_snap_wave=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_wave=rev_extrap_par%n_t_second
        rat_t_snap_wave=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave == 0) rat_t_snap_wave=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave=',min_it_snap_wave
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave=',rat_t_snap_wave
  
        min_it_snap_image=-1
        max_it_snap_image=-1
        rat_t_snap_image=-1

!        min_it_snap_image=nint((image_par%t_0-rev_extrap_par%t_0_second)/ &
!                              rev_extrap_par%d_t_second)
!        max_it_snap_image=rev_extrap_par%n_t_second
!        rat_t_snap_image=nint(image_par%d_t/rev_extrap_par%d_t_second)
!        if(rat_t_snap_image == 0) rat_t_snap_image=-1
!        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
!        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
        min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_data=rev_extrap_par%n_t_second-1
        rat_t_snap_data=nint(data_par%d_t/rev_extrap_par%d_t_second)

        if(rat_t_snap_data == 0) rat_t_snap_data=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data

      end if if_adj_2


    end if if_comp_domain_1 

    do_sy : do i_sy=1,wave_par%n_sy
    do_sx : do i_sx=1,wave_par%n_sx


    rev_extrap_par%i_sy=i_sy
    rev_extrap_par%i_sx=i_sx

    data_par%i_sy=rev_extrap_par%i_sy
    data_par%i_sx=rev_extrap_par%i_sx

    rev_extrap_par%i_t_extrap= 0

    if_comp_domain_2 : if (gen_par%comp_domain(1:4) == 'shot') then
      wave_par%sx=wave_par%sx_0 + (rev_extrap_par%i_sx-1)*wave_par%d_sx 
      wave_par%sy=wave_par%sy_0 + (rev_extrap_par%i_sy-1)*wave_par%d_sy 

      if(gen_par%verb) write(0,*)'i_sy=',i_sy,'at location y=',wave_par%sy
      if(gen_par%verb) write(0,*)'i_sx=',i_sx,'at location x=',wave_par%sx

      wave_par%mx_0=wave_par%sx + wave_par%off_x_0
      wave_par%my_0=wave_par%sy + wave_par%off_y_0
      if(gen_par%debug1) write(0,*)'wave_par-mx_0=',wave_par%mx_0
      if(gen_par%debug1) write(0,*)'wave_par-my_0=',wave_par%my_0

      wave_block=0.
    end if if_comp_domain_2 

    ! read appropriate veocity files
    if(slow_par%file == SLOW_INPUT) then
       if(gen_par%debug1) write(0,*)'Read Slow'
         call SEP_timer_start(11)
         call read_slow(gen_par,slow_par%tag, wave_par,slow_par, slow_block)
         call SEP_timer_stop(11)

    else
      if(gen_par%debug1) write(0,*)'DO NOT read Slow'
    end if


!go to 1000
    if(.not. gen_par%skip_source_prop) then
    !{
    if(((i_sx==1).and.(i_sy==1)).or.(.not. gen_par%layered_mig_vel)) then
    !{

    ! first propagate the source (for shot domain migration)
    if ((rev_extrap_par%adj).and.(gen_par%comp_domain(1:4) == 'shot')) then
    !{

      rev_extrap_par%i_t_extrap=0
      if(gen_par%verb) write(0,*)'(Source wavefield) Doing time step #',rev_extrap_par%i_t_extrap

      rev_extrap_par%wave_sou_step= &
       (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_sou,rat_t_snap_wave_sou)==0).and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_sou).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_sou)

      rev_extrap_par%wave_tmp_step= &
       (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_tmp,rat_t_snap_wave_tmp)==0).and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_tmp).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_tmp)

      if(rev_extrap_par%wave_tmp_step) then
        if(gen_par%debug1) write(0,*)'Output Wavefield_tmp'
        call output_wave(wave_par%tag_2,gen_par,wave_par,rev_extrap_par,wave_block)
      end if

      if(rev_extrap_par%wave_sou_step) then
        if(wave_par%file == WAVE_OUTPUT) then
          if(gen_par%debug1) write(0,*)'Output Wavefield_sou'
          call output_wave(wave_par%tag_1,gen_par,wave_par,rev_extrap_par,wave_block)
        end if
      end if

      do i_t_extrap=1,rev_extrap_par%n_t_first
      !{

        rev_extrap_par%i_t_extrap=i_t_extrap

        rev_extrap_par%t= rev_extrap_par%t_0_first+ &
                          (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_first

        if(gen_par%verb) write(0,*)'(Source wavefield) Doing time step #',rev_extrap_par%i_t_extrap
        if(gen_par%verb) write(0,*)'(Source wavefield) time=',rev_extrap_par%t


!        rev_extrap_par%wave_step= &
!            (mod(rev_extrap_par%i_t_extrap,rat_t_snap) == 0)

      rev_extrap_par%wave_sou_step= &
       (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_sou,rat_t_snap_wave_sou)==0).and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_sou).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_sou)

      rev_extrap_par%wave_tmp_step= &
       (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_tmp,rat_t_snap_wave_tmp)==0).and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_tmp).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_tmp)

!        rev_extrap_par%wave_tmp_step= &
!     ((abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) < epsi_image_d_t) &
!       .or. &
!      (image_par%d_t-abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) <  &
!             epsi_image_d_t) &
!     )

        call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                           force_block,wavelet)
  
  ! extrapolation step
  
        select case (gen_par%oper_type(1:2))
        case ('co')
          if(gen_par%debug1) write(0,*)'oper_type = Constant'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block &
              )
        case ('fu')
          if(gen_par%debug1) write(0,*)'oper_type = full'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block, &
              slow_block=slow_block)
        case default
           write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
           call seperr('Not supported oper_type')
        end select
  
        if(rev_extrap_par%wave_tmp_step) then
          if(gen_par%debug1) write(0,*)'Output Wavefield_tmp'
          call output_wave(wave_par%tag_2,gen_par,wave_par,rev_extrap_par,wave_block)
        end if
        if(wave_par%file == WAVE_OUTPUT) then
        !{
          if(rev_extrap_par%wave_sou_step) then
            if(gen_par%debug1) write(0,*)'Output Wavefield_sou'
            call output_wave(wave_par%tag_1,gen_par,wave_par,rev_extrap_par,wave_block)
          end if
        !}
        end if
      !}
      end do

      wave_block=0.

    !}
    end if

    !}
    end if
    !}
    end if

!1000 continue

! now propagate recorded wavefield and image

    rev_extrap_par%i_t_extrap=0
    rev_extrap_par%t= rev_extrap_par%t_0_second

    if(gen_par%verb) write(0,*)'Doing time step #',rev_extrap_par%i_t_extrap
    if(gen_par%verb) write(0,*)'time=',rev_extrap_par%t

    rev_extrap_par%wave_step= &
      (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave,rat_t_snap_wave) == 0).and.&
      (rev_extrap_par%i_t_extrap >= min_it_snap_wave).and.&
      (rev_extrap_par%i_t_extrap <= max_it_snap_wave)
    rev_extrap_par%data_step= &
      (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data) == 0).and.&
      (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
      (rev_extrap_par%i_t_extrap <= max_it_snap_data)

    if(rev_extrap_par%wave_step) then
      if(wave_par%file == WAVE_OUTPUT) then
        if(gen_par%debug1) write(0,*)'Output Wavefield'
        call output_wave(wave_par%tag,gen_par,wave_par,rev_extrap_par,wave_block)
      end if
    end if
  
    if(rev_extrap_par%data_step) then
      if(data_par%file == DATA_OUTPUT) then
      !{
        if(gen_par%debug1) write(0,*)'Output Data'
        call output_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par,&
                         wave_block)
      !}
      end if
    end if

    if (image_par%file == IMAGE_INPUT) then
    !{
      if(gen_par%debug1) write(0,*)'Input Image'
         call input_image(image_par%tag,gen_par,image_par, wave_par, &
                       wave_block)
    !}
    end if

!    !*$*assert do (serial) 
    do i_t_extrap=1,rev_extrap_par%n_t_second
    !{
      rev_extrap_par%i_t_extrap=i_t_extrap
      rev_extrap_par%t= rev_extrap_par%t_0_second+ &
         (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_second

      if(gen_par%verb) write(0,*)'Doing time step #',rev_extrap_par%i_t_extrap
      if(gen_par%verb) write(0,*)'time=',rev_extrap_par%t

      rev_extrap_par%wave_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave,rat_t_snap_wave)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_wave).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_wave)

      rev_extrap_par%image_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
        (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_image)

      rev_extrap_par%data_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_data)

!      rev_extrap_par%data_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-data_par%t_0,data_par%d_t)) < epsi_data_d_t) &
!     .or. &
!    (data_par%d_t-abs(amod(rev_extrap_par%t-data_par%t_0,data_par%d_t)) <  &
!           epsi_data_d_t) &
!   )
!
!      rev_extrap_par%wave_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-wave_par%t_0,wave_par%d_t)) < epsi_wave_d_t) &
!     .or. &
!    (wave_par%d_t-abs(amod(rev_extrap_par%t-wave_par%t_0,wave_par%d_t)) <  &
!           epsi_wave_d_t) &
!   )
!
!      rev_extrap_par%image_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) < epsi_image_d_t) &
!     .or. &
!    (image_par%d_t-abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) <  &
!           epsi_image_d_t) &
!   )

      if (data_par%file == DATA_INPUT) then
        !{
        if(rev_extrap_par%data_step ) then
          if(gen_par%debug1) write(0,*)'Read Data'
          call SEP_timer_start(14)
          call input_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par, &
                           force_block,wave_block)
          call SEP_timer_stop(14)
        end if
        !}
      end if


      if (.not. rev_extrap_par%adj) then
        if(wavelet_par%file == WAVELET_INPUT) then
          call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                           force_block,wavelet)
        end if
      end if

! extrapolation step

      select case (gen_par%oper_type(1:2))
      case ('co')
        if(gen_par%debug1) write(0,*)'oper_type = Constant'
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block &
            )
      case ('fu')
        if(gen_par%debug1) write(0,*)'oper_type = full'
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block, &
            slow_block=slow_block)
      case default
         write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
         call seperr('Not supported oper_type')
      end select

      if(wave_par%file == WAVE_OUTPUT) then
      !{
        if(rev_extrap_par%wave_step) then
          if(gen_par%debug1) write(0,*)'Output Wavefield'
          call output_wave(wave_par%tag,gen_par,wave_par,rev_extrap_par,wave_block)
        end if
      !}
      end if

      if(data_par%file == DATA_OUTPUT) then
      !{
        if(rev_extrap_par%data_step ) then
          if(gen_par%debug1) write(0,*)'Output Data'
          call output_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par,&
                         wave_block)
       end if
      !}
      end if

      if (image_par%file == IMAGE_OUTPUT) then
      !{
        if(rev_extrap_par%image_step ) then
          if(gen_par%debug1) write(0,*)'Output Image'
          call output_image(image_par%tag,image_par%tag_hz,gen_par,image_par, &
                             wave_par,rev_extrap_par, &
                             max_it_snap_image, &
                             wave_block)
        end if
      !}
      end if


    !}
    end do


    end do do_sx 
    end do do_sy 

    stat=0
    if(any(stat_par /= 0)) stat =1
    return
  end function rev_extrap_gen
end module rev_extrap_gen_mod
