module wave_mod

  use genrev3d_types
  use wave_types
  use slow_types
  use rev_extrap_types


  implicit none

  logical, private :: adj,add

  integer, external ::  sreed,srite,sseek_block
  integer, private :: n_bytes_io,n_ztau

  real, allocatable, dimension (:,:,:), private :: wave
  
contains

  !-----------------------------------------------------------------------------
  subroutine init_io_wave(gen_par,wave_par)

    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    integer ierr

    !+++
    if(gen_par%tau_domain) then
      n_ztau=wave_par%n_tau
      n_bytes_io=n_ztau*wave_par%n_mx* &
                 wave_par%n_my*wave_par%esize
    else
      n_ztau=wave_par%n_z
      n_bytes_io=n_ztau*wave_par%n_mx* &
                 wave_par%n_my*wave_par%esize
    end if

    allocate (wave(n_ztau,wave_par%n_mx,wave_par%n_my),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of wave in init_io_wave")

  end subroutine init_io_wave


  !-----------------------------------------------------------------------------
  subroutine output_wave(tag_wave,gen_par,wave_par,rev_extrap_par,wave_block)

   character(len=128) :: tag_wave
   type (genrev3d_param_type) gen_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) :: rev_extrap_par
   real,  dimension (:,:,:,:,:,:)   :: wave_block

   logical :: rewind
   integer :: len_tag,i_hx,i_hy,ret_sseek

   len_tag=len_trim(tag_wave)

   rewind = ((tag_wave(1:len_tag) == wave_par%tag_2(1:len_tag)) .and. &
             (rev_extrap_par%i_t_extrap == 1) .and.  &
             ((rev_extrap_par%i_sx /= 1) .or. (rev_extrap_par%i_sy /= 1)) &
            )

  write(0,*)'rewind=',rewind

!write(0,*)'n_ztau=',n_ztau
!write(0,*)'n_my=',wave_par%n_my
!write(0,*)'n_mx=',wave_par%n_mx
!write(0,*)'n_hy=',wave_par%n_hy
!write(0,*)'n_hx=',wave_par%n_hx

!   if(gen_par%debug2) write(0,*)'len_tag',len_tag
!
!   if(gen_par%debug2) write(0,*)'Writing to ',tag_wave(1:len_tag)
!
!   if(gen_par%debug2) write(0,*)'shape(wave)',shape(wave)
!   if(gen_par%debug2) write(0,*)'shape(wave_block)',shape(wave_block)


   if(rewind) then
   ret_sseek=sseek_block(trim(tag_wave),0,n_bytes_io,0) 
     if (ret_sseek .ne. 0) then
     !{
       write(0,*)'ret_sseek=',ret_sseek
       write(0,*)'tag_wave=',trim(tag_wave)
       call seperr("Error block seeking wave in output")
     !}
     end if
   end if

   do i_hy=1,wave_par%n_hy
     do i_hx=1,wave_par%n_hx
       wave=wave_block(wave_par%i_t_step(1),:,:,:,i_hx,i_hy)
       if(srite(trim(tag_wave),wave,n_bytes_io) .ne. n_bytes_io ) then
         write(0,*)'tag_wave=',trim(tag_wave)
         write(0,*)'i_hx=',i_hx
         write(0,*)'i_hy=',i_hy
         call seperr('Problems with writing tag_wave file')
       end if
     end do
   end do


  return	
  end subroutine output_wave


  !-----------------------------------------------------------------------------
  subroutine input_wave(tag_wave,gen_par,wave_par,wave_block)

   character(len=128) :: tag_wave
   type (genrev3d_param_type) gen_par
   type (wave_param_type) wave_par
   real, dimension (:,:,:,:,:,:)   :: wave_block

   integer :: i_hx,i_hy
   real   :: wave(n_ztau,wave_par%n_mx,wave_par%n_my)


   do i_hy=1,wave_par%n_hy
     do i_hx=1,wave_par%n_hx
       if(sreed(trim(tag_wave),wave,n_bytes_io) .ne. n_bytes_io ) then
         write(0,*)'tag_wave=',trim(tag_wave)
         write(0,*)'i_hx=',i_hx
         write(0,*)'i_hy=',i_hy
         call seperr('Problems with writing tag_wave file')
       end if
       wave_block(wave_par%i_t_step(1),:,:,:,i_hx,i_hy)=wave
     end do
   end do


   return	
   end subroutine input_wave

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_wave()
    integer ierr

    if(allocated(wave)) deallocate (wave,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of wave in clean_io_wave")

    return
  end subroutine clean_io_wave


end module wave_mod
