module Rev2D_zoff_mod

  use Rev2D_zoff_types

  use genrev3d_types
  use rev_extrap_types
  use data_types
  use slow_types
  use wave_types

  use fft_data_mod
!  use fft_shan_mod

  implicit none

  integer, private:: n_mx_shan,n_z_shan

  type (Rev2D_zoff_param_type), private :: Rev2D_zoff_par_loc

  integer, private ::i_ztau_sou
  integer, private ::nop

  real, allocatable, dimension (:), private :: k_mx,k_z
  complex, allocatable, dimension (:,:), private :: temp
  complex, allocatable, dimension (:,:), private :: temp_mx,temp_z
  real, allocatable, save, dimension(:,:,:,:) ::q_x,q_z !i_tm1,qi_tm2,qi_t0
  real, dimension(:), pointer                 ::xtaper,ztaper
  real                                        ::xtapco,ztapco 
  integer,private                             ::n_btz,n_btx
contains

  subroutine init_Rev2D_zoff_par(Rev2D_zoff_par)
    type (Rev2D_zoff_param_type) Rev2D_zoff_par
    Rev2D_zoff_par_loc=Rev2D_zoff_par
  end subroutine init_Rev2D_zoff_par

  function Rev2D_zoff_init(gen_par,data_par,slow_par,wave_par)  result(info)
    type (genrev3d_info_type) info
    type (genrev3d_param_type) gen_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par
    type (wave_param_type) wave_par

    integer i_k_mx,i_k_hx,i_k_z,ierr,i_z,i_mx,n_z,m_mx_pos
    real kz,kmx,khx,epsi

    info=info_default

    wave_par%n_t_step=3
    nop=5

    n_btz=20
    n_btx=20
    xtapco=0.0003
    ztapco=0.0003

    n_mx_shan=156
    n_z_shan=208
    n_mx_shan=wave_par%n_mx
    n_z_shan=wave_par%n_z
    

    if(gen_par%tau_domain) then
      i_ztau_sou=(gen_par%sou_depth-wave_par%tau_0)/wave_par%d_tau+1
    else
      i_ztau_sou=(gen_par%sou_depth-wave_par%z_0)/wave_par%d_z+1
    end if
    allocate (q_x(wave_par%n_z,nop,3,2), stat=ierr) 
    ! 3 mean qi_t0,qi_tm1,qi_tm2;  2 mean x_max and x_min 
    if(ierr .ne. 0) call seperr("Problems with allocation of qx")

    allocate (q_z(wave_par%n_mx,nop,3,2), stat=ierr) 
    ! 3 mean qi_t0,qi_tm1,qi_tm2;  2 mean z_max and z_min 
    if(ierr .ne. 0) call seperr("Problems with allocation of qz")
    allocate(ztaper(n_btz),stat=ierr)
    if(ierr .ne. 0) call seperr("Problems with allocation of ztaper")
    allocate(xtaper(n_btx),stat=ierr)
    if(ierr .ne. 0) call seperr("Problems with allocation of xtaper")

    q_x=0.0
    q_z=0.0
    do i_z=1,n_btz
       ztaper(i_z)=exp(-ztapco*i_z*i_z)
    end do
    do i_mx=1,n_btx
       xtaper(i_mx)=exp(-xtapco*i_mx*i_mx)
    end do

    allocate (temp(n_z_shan,n_mx_shan), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp ")
    allocate (temp_z(n_z_shan,gen_par%n_threads), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_z ")
    allocate (temp_mx(n_mx_shan,gen_par%n_threads), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_mx ")

    allocate (k_mx(n_mx_shan),stat=ierr)
    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
    allocate (k_z(n_z_shan),stat=ierr)
    if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
    do i_k_mx=1,n_mx_shan
       k_mx(i_k_mx)=(-1.0+(i_k_mx-1)*2.0/n_mx_shan)*3.1415926/wave_par%d_mx
    end do
    do i_k_z=1,n_z_shan
       k_z(i_k_z)=(-1.0+(i_k_z-1)*2.0/n_z_shan)*3.1415926/wave_par%d_z
    end do
 
  end function Rev2D_zoff_init

  function Rev2D_zoff_extrap_oper( &
    rev_extrap_par, &
    wave_par, &
    force_block, &
    wave_block,slow_block) &
    result(stat)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx,i_kmx,ierr
    integer i_t0,i_tm1,i_tm2
    integer n_threads,i_thread
    real vel_sq,kmx_sq,kz_sq
    real d_t_sq
    real der2
!    real f

    optional slow_block

    real,  dimension (:,:,:,:)      :: force_block
    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block
    real bndz0,delp,vel,d_tz,d_tx,bndx0,bndx1
    integer iop,ib_t0,ib_tm1,ib_tm2,b_min,b_max

    ib_t0=3
    ib_tm1=2
    ib_tm2=1
    b_min=1
    b_max=2

     
    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    d_t_sq=(rev_extrap_par%d_t)**2
    d_tz=rev_extrap_par%d_t/wave_par%d_z
    d_tx=rev_extrap_par%d_t/wave_par%d_mx


    i_t0=wave_par%i_t_step(1)
    i_tm1=wave_par%i_t_step(2)
    i_tm2=wave_par%i_t_step(3)
    
    n_z=wave_par%n_z
    n_mx=wave_par%n_mx
    
    if (rev_extrap_par%i_t_extrap == 1) then
       q_x=0.0
       q_z=0.0
    end if
!    do i_z=1,n_btz
!       f=(i_z+n_btz)/(2.0*n_btz)
!       slow_block(i_z,:,1,1,1,1) =f*slow_block(1+n_btz,:,1,1,1,1)+(1-f)*slow_block(n_z-n_btz,:,1,1,1,1)
!       slow_block(n_z-i_z+1,:,1,1,1,1)=(1-f)*slow_block(1+n_btz,:,1,1,1,1)+f*slow_block(n_z-n_btz,:,1,1,1,1)
!    end do

![[[[[[[[[[[[[[[[[[[[[[[[[ laplac in Fourier domain

!$omp parallel default (private), shared(wave_block,temp_mx,temp), &
!$omp& shared(n_z_shan,n_mx_shan,n_threads,i_t0)
!$omp do      
      do i_thread=1,n_threads  
         do i_z=i_thread,n_z_shan,n_threads
           do i_mx=1,n_mx_shan
             temp_mx(i_mx,i_thread)=cmplx(wave_block(i_t0,i_z,i_mx,1,1,1))
           end do 
           call segfft1d(temp_mx(:,i_thread),n_mx_shan,1)
           do i_mx=1,n_mx_shan
             temp(i_z,i_mx)=temp_mx(i_mx,i_thread)
           end do 
         end do
      end do
!$omp end parallel


!$omp parallel default (private),shared (temp,temp_z,k_mx,k_z), &
!$omp& shared(n_mx_shan,n_z_shan,n_threads)
!$omp do
      do i_thread=1,n_threads
        do i_mx=i_thread,n_mx_shan,n_threads
          kmx_sq=k_mx(i_mx)**2
          do i_z=1,n_z_shan
            temp_z(i_z,i_thread)=temp(i_z,i_mx)
          end do 
          call segfft1d(temp_z(:,i_thread),n_z_shan,1)
          do i_z=1,n_z_shan
            kz_sq=k_z(i_z)**2
            der2=-(kmx_sq + kz_sq)
            temp_z(i_z,i_thread)=temp_z(i_z,i_thread)*der2
          end do 
          call segfft1d(temp_z(:,i_thread),n_z_shan,-1)
          do i_z=1,n_z_shan
            temp(i_z,i_mx)=temp_z(i_z,i_thread)
          end do 
        end do 
      end do
!$omp end parallel

!$omp parallel default (private), shared (temp,temp_mx), &
!$omp& shared(n_z_shan,n_mx_shan,n_threads)
!$omp do
      do i_thread=1,n_threads
        do i_z=i_thread,n_z_shan,n_threads
          do i_mx=1,n_mx_shan
            temp_mx(i_mx,i_thread)=temp(i_z,i_mx)
          end do 
          call segfft1d(temp_mx(:,i_thread),n_mx_shan,-1)
          do i_mx=1,n_mx_shan
            temp(i_z,i_mx)=temp_mx(i_mx,i_thread)
          end do 
        end do 
      end do
!$omp end parallel


!      end laplac in Fourier domain ]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]]


!!{{{{{{{{{{{{{{{{{{{{{{{{{  taper at boundary
      do i_z=1,n_z
         wave_block(i_tm1,i_z,1:n_btx,1,1,1)= &
          wave_block(i_tm1,i_z,1:n_btx,1,1,1)*xtaper(n_btx:1:-1)
         wave_block(i_tm1,i_z,n_mx-n_btx+1:n_mx)= &
          wave_block(i_tm1,i_z,n_mx-n_btx+1:n_mx)*xtaper(1:n_btx:+1)
         wave_block(i_t0,i_z,1:n_btx,1,1,1)= &
          wave_block(i_t0,i_z,1:n_btx,1,1,1)*xtaper(n_btx:1:-1)
         wave_block(i_t0,i_z,n_mx-n_btx+1:n_mx)= &
          wave_block(i_t0,i_z,n_mx-n_btx+1:n_mx)*xtaper(1:n_btx:+1)
         temp(i_z,1:n_btx)=temp(i_z,1:n_btx)*xtaper(n_btx:1:-1)
         temp(i_z,n_mx-n_btx+1:n_mx)=temp(i_z,n_mx-n_btx+1:n_mx)*xtaper(1:n_btx:+1)
      end do
      do i_mx=1+n_btx,n_mx-n_btx
         wave_block(i_tm1,1:n_btz,i_mx,1,1,1)= &
          wave_block(i_tm1,1:n_btz,i_mx,1,1,1)*ztaper(n_btz:1:-1)
         wave_block(i_tm1,n_z-n_btz+1:n_z,i_mx,1,1,1)= &  
          wave_block(i_tm1,n_z-n_btz+1:n_z,i_mx,1,1,1)*ztaper(1:n_btz:+1)
        wave_block(i_t0,1:n_btz,i_mx,1,1,1)= &
          wave_block(i_t0,1:n_btz,i_mx,1,1,1)*ztaper(n_btz:1:-1)
         wave_block(i_t0,n_z-n_btz+1:n_z,i_mx,1,1,1)= &  
          wave_block(i_t0,n_z-n_btz+1:n_z,i_mx,1,1,1)*ztaper(1:n_btz:+1)
          temp(1:n_btz,i_mx)=temp(1:n_btz,i_mx)*ztaper(n_btz:1:-1)
          temp(n_z-n_btz+1:n_z,i_mx)=temp(n_z-n_btz+1:n_z,i_mx)*ztaper(1:n_btz:+1)
      end do
!                        end taper at boundary}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
              

!     ((((((((((((((((((((((extroplation of wave field
      do i_mx=1,n_mx
         do i_z=1,n_z
          wave_block(i_tm2,i_z,i_mx,1,1,1)= &
              wave_block(i_tm1,i_z,i_mx,1,1,1)
          wave_block(i_tm1,i_z,i_mx,1,1,1)= &
              wave_block(i_t0,i_z,i_mx,1,1,1)            
         end do
      end do


      do i_mx=1,n_mx
        do i_z=1,n_z
          vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)
          wave_block(i_t0,i_z,i_mx,1,1,1)= &
            2*wave_block(i_tm1,i_z,i_mx,1,1,1) - &
              wave_block(i_tm2,i_z,i_mx,1,1,1) + &
              d_t_sq*(vel_sq*real(temp(i_z,i_mx)))
        end do
      end do 
!     end extroplation wave field)))))))))))))))))))))) 

!!!!!!!!!!! solve auxiliary wavefield  Q !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   q_x(:,:,ib_tm2,:)=q_x(:,:,ib_tm1,:)
   q_x(:,:,ib_tm1,:)=q_x(:,:,ib_t0,:)
   q_x(1,:,ib_t0,:)=0.0
   q_x(n_z,:,ib_t0,:)=0.0

   do iop=1,nop
      do i_z=2,n_z-1
         vel=1.0/slow_block(i_z,iop,1,1,1,1)
         delp=vel*d_tz
         q_x(i_z,iop,ib_t0,b_min)=   &
    +2.0*q_x(i_z,iop,ib_tm1,b_min)   & 
    -    q_x(i_z,iop,ib_tm2,b_min)   &
    -    0.5*0.4761*delp*delp*( &
    +    wave_block(i_tm1,i_z+1,iop,  1,1,1)      &
    -2.0*wave_block(i_tm1,i_z  ,iop,  1,1,1)      &
    +    wave_block(i_tm1,i_z-1,iop,  1,1,1)      &
    +    wave_block(i_tm1,i_z+1,iop+1,1,1,1)      &
    -2.0*wave_block(i_tm1,i_z  ,iop+1,1,1,1)      &
    +    wave_block(i_tm1,i_z-1,iop+1,1,1,1)   )  &
    +    0.3741*delp*delp*( &
    +    q_x(i_z+1,iop,ib_tm1,b_min) &
    -2.0*q_x(i_z  ,iop,ib_tm1,b_min)   &
    +    q_x(i_z-1,iop,ib_tm1,b_min)  )
      end do
   end do

   do iop=1,nop
      do i_z=2,n_z-1
         vel=1.0/slow_block(i_z,n_mx-nop+iop,1,1,1,1)
         delp=vel*d_tz
         q_x(i_z,iop,ib_t0,b_max)= &
    +2.0*q_x(i_z,iop,ib_tm1,b_max) &
    -    q_x(i_z,iop,ib_tm2,b_max) &
    -    0.5*0.4761*delp*delp*( &
    +    wave_block(i_tm1,i_z+1,n_mx-nop+iop,  1,1,1)     &
    -2.0*wave_block(i_tm1,i_z  ,n_mx-nop+iop,  1,1,1)     &
    +    wave_block(i_tm1,i_z-1,n_mx-nop+iop,  1,1,1)     &
    +    wave_block(i_tm1,i_z+1,n_mx-nop+iop-1,1,1,1)     &
    -2.0*wave_block(i_tm1,i_z  ,n_mx-nop+iop-1,1,1,1)     &
    +    wave_block(i_tm1,i_z-1,n_mx-nop+iop-1,1,1,1)   ) &
    +0.3741*delp*delp*(  &
    +    q_x(i_z+1,iop,ib_tm1,b_max) &
    -2.0*q_x(i_z  ,iop,ib_tm1,b_max) &
    +    q_x(i_z-1,iop,ib_tm1,b_max)  &
                       )
      end do
   end do

   q_z(:,:,ib_tm2,:)=q_z(:,:,ib_tm1,:)
   q_z(:,:,ib_tm1,:)=q_z(:,:,ib_t0 ,:)
   q_z(1,   :,ib_t0,:)=0.0
   q_z(n_mx,:,ib_t0,:)=0.0

   do iop=1,nop
      do i_mx=2,n_mx-1
         vel=1.0/slow_block(iop,i_mx,1,1,1,1)
         delp=vel*d_tx
         q_z(i_mx,iop,ib_t0,b_min)=   &
    +2.0*q_z(i_mx,iop,ib_tm1,b_min)   & 
    -    q_z(i_mx,iop,ib_tm2,b_min)   &
    -    0.5*0.4761*delp*delp*( &
    +    wave_block(i_tm1,iop,  i_mx+1,1,1,1)      &
    -2.0*wave_block(i_tm1,iop,  i_mx  ,1,1,1)      &
    +    wave_block(i_tm1,iop,  i_mx-1,1,1,1)      &
    +    wave_block(i_tm1,iop+1,i_mx+1,1,1,1)      &
    -2.0*wave_block(i_tm1,iop+1,i_mx  ,1,1,1)      &
    +    wave_block(i_tm1,iop+1,i_mx-1,1,1,1)   )  &
    +    0.3741*delp*delp*( &
    +    q_z(i_mx+1,iop,ib_tm1,b_min) &
    -2.0*q_z(i_mx  ,iop,ib_tm1,b_min)   &
    +    q_z(i_mx-1,iop,ib_tm1,b_min)  )
      end do
   end do

   do iop=1,nop
      do i_mx=2,n_mx-1
         vel=1.0/slow_block(n_z-nop+iop,i_mx,1,1,1,1)
         delp=vel*d_tx
         q_z(i_mx,iop,ib_t0,b_max)= &
    +2.0*q_z(i_mx,iop,ib_tm1,b_max) &
    -    q_z(i_mx,iop,ib_tm2,b_max) &
    -    0.5*0.4761*delp*delp*( &
    +    wave_block(i_tm1,n_z-nop+iop,i_mx+1,  1,1,1)     &
    -2.0*wave_block(i_tm1,n_z-nop+iop,i_mx  ,  1,1,1)     &
    +    wave_block(i_tm1,n_z-nop+iop,i_mx-1,  1,1,1)     &
    +    wave_block(i_tm1,n_z-nop+iop-1,i_mx+1,1,1,1)     &
    -2.0*wave_block(i_tm1,n_z-nop+iop-1,i_mx,  1,1,1)     &
    +    wave_block(i_tm1,n_z-nop+iop-1,i_mx-1,1,1,1)   ) &
    +0.3741*delp*delp*(  &
    +    q_z(i_mx+1,iop,ib_tm1,b_max) &
    -2.0*q_z(i_mx  ,iop,ib_tm1,b_max) &
    +    q_z(i_mx-1,iop,ib_tm1,b_max)  )
      end do
   end do
!!!!!!!!! above  solve the auxiliary wavefied Q !!!!!!!!!!!   


!@@@@@@@@@@@@@@@@@@@@@@@ one way wave equation boundary condition@@@@@@@@@@@@@@
  do iop=nop,1,-1
     do i_mx=1,n_mx
        vel=1.0/slow_block(iop,i_mx,1,1,1,1)
        delp=vel*d_tz
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(i_t0,iop,i_mx,1,1,1)= &
                   wave_block(i_tm1,iop+1,i_mx,1,1,1)+ &
             bndz0*(wave_block(i_tm1,iop,i_mx,1,1,1)- &
              wave_block(i_t0,iop+1,i_mx,1,1,1) )+ &
           2.0/(1+delp)*( q_z(i_mx,iop,ib_tm1,b_min)-q_z(i_mx,iop,ib_t0,b_min) ) 
     end do
     do i_mx=1,n_mx
        vel=1.0/slow_block(n_z-iop+1,i_mx,1,1,1,1)
        delp=vel*d_tz
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(i_t0,n_z-iop+1,i_mx,1,1,1)= &
                    wave_block(i_tm1,n_z-iop,i_mx,1,1,1)+ &
             bndz0*(wave_block(i_tm1,n_z-iop+1,i_mx,1,1,1)- &
              wave_block(i_t0,n_z-iop,i_mx,1,1,1) )+ &
          2.0/(1+delp)*( q_z(i_mx,nop-iop+1,ib_tm1,b_max)-q_z(i_mx,nop-iop+1,ib_t0,b_max) ) 
     end do
     do i_z=1+nop,n_z-nop
        vel=1.0/slow_block(i_z,iop,1,1,1,1)
        delp=vel*d_tx
        bndx0= (1.-delp)/(1.+delp)
        wave_block(i_t0,i_z,iop,1,1,1)=&
           wave_block(i_tm1,i_z,iop+1,1,1,1)+ &
           bndx0*(wave_block(i_tm1,i_z,iop,1,1,1)- &     
                wave_block(i_t0,i_z,iop+1,1,1,1) )   + &
            2.0/(1+delp)*( q_x(i_z,iop,ib_tm1,b_min)-q_x(i_z,iop,ib_t0,b_min) )     
     end do

     do i_z=1+nop,n_z-nop
        vel=1.0/slow_block(i_z,n_mx-iop+1,1,1,1,1)
        delp = vel*d_tx
        bndx1= (1.-delp)/(1.+delp) 
        wave_block(i_t0,i_z,n_mx-iop+1,1,1,1)=&
            wave_block(i_tm1,i_z,n_mx-iop,1,1,1)+ &   
            bndx1*(wave_block(i_tm1,i_z,n_mx-iop+1,1,1,1)- &
                  wave_block(i_t0,i_z,n_mx-iop,1,1,1) ) +  &
            2.0/(1+delp)*( q_x(i_z,nop-iop+1,ib_tm1,b_max)-q_x(i_z,nop-iop+1,ib_t0,b_max)  ) 
     end do

  end do
!@@@@@@@@@@@@@@@@@@@@@@@ end of boundary condition @@@@@@@@@@@@@@@@@@@@@@@@@@

!/////////////////////  source //////////////////////////////////////////////
  do i_mx=1,n_mx
     wave_block(i_t0,i_ztau_sou,i_mx,1,1,1)=&
         wave_block(i_t0,i_ztau_sou,i_mx,1,1,1) &
         +d_t_sq*force_block(i_mx,1,1,1)
  end do    
!///////////////////   end source  /////////////////////////////////////////


    stat=0
    return	
  end function Rev2D_zoff_extrap_oper


end module Rev2D_zoff_mod
